ExtraSpellAlert = LibStub("AceAddon-3.0"):NewAddon("ExtraSpellAlert", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0");
local core = ExtraSpellAlert;
local L = LibStub("AceLocale-3.0"):GetLocale("ExtraSpellAlert")

local options = {
	name = "ExtraSpellAlert",
	type = 'group',
	args = {
		toggle = {
			name = L["Toggle Individual Spells"],
			type = 'group',
			order = 0,
			inline = true,
			args = {
				desc = {
					name = L["Spells will be added to this list once they have been gained at least once"],
					type = 'description',
					order = 0,
				},
				-- loaded dynamically
			},
		},
		stacks = {
			name = L["Animate Buff Charges"],
			type = 'group',
			order = 10,
			inline = true,
			args = {
				enabled = {
					name = L["Enabled"],
					type = 'toggle',
					get = function(info) return core.db.profile.stacks.enabled end,
					set = function(info,v) core.db.profile.stacks.enabled = v end,
				},
			},
		},
		warn = {
			name = L["Low Buff Warning"],
			type = 'group',
			order = 20,
			inline = true,
			args = {
				enabled = {
					name = L["Enabled"],
					type = 'toggle',
					order = 0,
					get = function(info) return core.db.profile.warn.enabled end,
					set = function(info,v) core.db.profile.warn.enabled = v end,
				},
				thresh = {
					name = L["Time Threshhold"],
					type = 'range',
					order = 20,
					min = 1,
					max = 30,
					step = 1,
					get = function(info) return core.db.profile.warn.thresh end,
					set = function(info,v) core.db.profile.warn.thresh = v end,
				},
				color = {
					name = L["Color Tint"],
					type = 'color',
					order = 10,
					hasAlpha = false,
					get = function(info) 
						return core.db.profile.warn.color.r, core.db.profile.warn.color.g, core.db.profile.warn.color.b, 1
					end,
					set = function(info,r,g,b) 
						core.db.profile.warn.color.r = r
						core.db.profile.warn.color.g = g
						core.db.profile.warn.color.b = b
					end,
				},
			},
		},
	},
};
local defaults = {
	profile = {
		stacks = {
			enabled = true,
		},
		warn = {
			enabled = true,
			thresh = 5,
			color = {
				r = 1,
				g = 0.25,
				b = 0.25,
			},
		},
		toggle = {
		},
	},
};

--[[ TODO
	Toggle specific auras from showing
	Tint red (as animation?) with low time
]]

local saoFrame = SpellActivationOverlayFrame
local dbBuff = { }; -- populated by SPELL_ACTIVATION_OVERLAY_SHOW

-- Blizzard locals
local sizeScale = 0.8;
local longSide = 256 * sizeScale;
local shortSide = 128 * sizeScale;

core.debug = 0;
--[===[@debug@
core.debug = 1; -- debug printing by level
--@end-debug@]===]
core.buffStatus = {  -- populated by SPELL_ACTIVATION_OVERLAY_SHOW

	--[[[44544] = {			-- Fingers of Frost
		count = 0,
		redFlag = false,
	},]]
	
};
core.overlayList = { }; -- registry of all overlays

function core:BuildOptions()
	for i,v in pairs(self.db.profile.toggle) do
		options.args.toggle.args[tostring(i)] = {
			name = GetSpellInfo(i),
			type = 'toggle',
			order = i,
			get = function(info) return core.db.profile.toggle[i] end,
			set = function(info,v) core.db.profile.toggle[i] = v end,
		};
		--[[options.args.toggle.args["br-"..i] = {
			name = "",
			type = 'description',
			order = i+0.5,
		};]]
	end
end

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ExtraSpellAlertDB", defaults, "Default");
	-- populate table
	self:BuildOptions();
	
	self:RegisterChatCommand("esa", "SlashHandle");
	self:RegisterChatCommand("esalert", "SlashHandle");
	self:RegisterChatCommand("extraspellalert", "SlashHandle");
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("ExtraSpellAlert", options);
	local ACD = LibStub("AceConfigDialog-3.0");
	ACD:AddToBlizOptions("ExtraSpellAlert", "ExtraSpellAlert");
end

function core:OnEnable()
	self:RegisterEvent("SPELL_ACTIVATION_OVERLAY_SHOW");
	self:RegisterEvent("UNIT_AURA");
	self:RawHook("SpellActivationOverlay_ShowAllOverlays", true);
end

function core:OnDisable()
	
end

function core:SlashHandle()
	LibStub("AceConfigDialog-3.0"):Open("ExtraSpellAlert")
end

function core:SPELL_ACTIVATION_OVERLAY_SHOW(_, spellID, texture, positions, scale, r, g, b)
	-- cache information
	if (not dbBuff[spellID]) then
		dbBuff[spellID] = {
			texture = texture,
			positions = positions,
			scale = scale,
			r = r,
			g = g,
			b = b,		
		};
		-- refire UNIT_AURA in case we simply didn't have a cache before
		self:UNIT_AURA(nil, "player");
	end
	
	-- create status
	if (not self.buffStatus[spellID]) then
		self.buffStatus[spellID] = {
			count = 0,
			redFlag = false,
		};
	end
	
	-- check toggle cache
	if (self.db.profile.toggle[spellID] == nil) then
		self.db.profile.toggle[spellID] = true
		self:BuildOptions();
	end	
	
	-- schedule a check for low
	self:CheckTime(spellID);
end

function core:UNIT_AURA(_, unit)
	-- check if activations are enabled
	if (not GetCVarBool("displaySpellActivationOverlays")) then
		return
	end
	
	-- check if frame is enabled
	if (not saoFrame) then
		if (not SpellActivationOverlayFrame) then
			return;
		end
		saoFrame = SpellActivationOverlayFrame;
	end	
	
	if (unit == "player") then
		for i,v in pairs(self.buffStatus) do	
			-- check that we have db info
			if (dbBuff[i]) then
				local _, _, _, count, _, _, expires = self:UnitAura(GetSpellInfo(i));
				
				-- check if any stacks changed
				count = count or 0;
				-- self:Print("UNIT_AURA checking " .. i .. " with " .. (count or 0) .. " stacks (current: " .. v .. ")");
				if (v.count < count) then
					-- increase in stacks; max() prevents creating one at 1 stack
					for j=max(2, v.count+1),count do 
						self:ShowOverlay(j, i, dbBuff[i].texture, string.upper(dbBuff[i].positions), dbBuff[i].scale*(1+0.2*(j-1)), dbBuff[i].r, dbBuff[i].g, dbBuff[i].b, nil, nil, 1-0.2*(j-1));
					end
				elseif (v.count > count) then
					-- decrease in stacks; hide extras
					for j=count+1,v.count do
						self:HideOverlay(j, i);
					end
				end
				self.buffStatus[i].count = count;
				
				if (v.redFlag and expires and expires-GetTime()>self.db.profile.warn.thresh) then 
					-- buff was refreshed
					self:CheckTime(i);
					self.redFlag = false;
					SpellActivationOverlay_ShowAllOverlays(saoFrame, i, dbBuff[i].texture, dbBuff[i].positions, dbBuff[i].scale, dbBuff[i].r, dbBuff[i].g, dbBuff[i].b);
				end
			end
		end
	end
end

-- functions similar to the blizzard functions in SpellActivationOverlay.lua
-- need to hardcode these myself because they only allow one animation per position ("TOP")

function core:ShowOverlay(index, spellID, texturePath, position, scale, r, g, b, vFlip, hFlip, alpha)
	-- self:Print(index, spellID, texturePath, position, scale, r, g, b, vFlip, hFlip);
	local overlay = self:GetUnusedOverlay();
	overlay.esaIndex = index;
	overlay.esaSpellID = spellID;
	self.overlayList[overlay] = true;
	
	overlay:SetFrameLevel(1);
	if (alpha and alpha < 1) then
		overlay.animIn:GetAnimations():SetChange(alpha);
		overlay.animIn:SetScript("OnFinished", function(anim)
			local overlay = anim:GetParent();
			overlay:SetAlpha(alpha);
			overlay.pulse:Play();
		end);
	end
	
	if (overlay.animOut:GetAnimations():HasScript("OnFinished")) then
		-- due to a FrameXML inconsistency on SpellActivationOverlay.xml:25 I need to traverse into the animation to set the script
		overlay.animOut:GetAnimations():SetScript("OnFinished", function(anim)
			self:OnFadeOutFinished(anim);
		end)
	else
		-- the inconsistency was fixed sometime
		overlay.animOut:SetScript("OnFinished")
	end
	
	
	-- blizzard
	overlay.spelID = spellID;
	overlay.position = position;
	
	overlay:ClearAllPoints();
	local texLeft, texRight, texTop, texBottom = 0, 1, 0, 1;
	if ( vFlip ) then
		texTop, texBottom = 1, 0;
	end
	if ( hFlip ) then
		texLeft, texRight = 1, 0;
	end
	overlay.texture:SetTexCoord(texLeft, texRight, texTop, texBottom);
	
	local width, height;
	if ( position == "CENTER" ) then
		width, height = longSide, longSide;
		overlay:SetPoint("CENTER", saoFrame, "CENTER", 0, 0);
	elseif ( position == "LEFT" ) then
		width, height = shortSide, longSide;
		overlay:SetPoint("RIGHT", saoFrame, "LEFT", 0, 0);
	elseif ( position == "RIGHT" ) then
		width, height = shortSide, longSide;
		overlay:SetPoint("LEFT", saoFrame, "RIGHT", 0, 0);
	elseif ( position == "TOP" ) then
		width, height = longSide, shortSide;
		overlay:SetPoint("BOTTOM", saoFrame, "TOP");
	elseif ( position == "BOTTOM" ) then
		width, height = longSide, shortSide;
		overlay:SetPoint("TOP", saoFrame, "BOTTOM");
	elseif ( position == "TOPRIGHT" ) then
		width, height = shortSide, shortSide;
		overlay:SetPoint("BOTTOMLEFT", saoFrame, "TOPRIGHT", 0, 0);
	elseif ( position == "TOPLEFT" ) then
		width, height = shortSide, shortSide;
		overlay:SetPoint("BOTTOMRIGHT", saoFrame, "TOPLEFT", 0, 0);
	elseif ( position == "BOTTOMRIGHT" ) then
		width, height = shortSide, shortSide;
		overlay:SetPoint("TOPLEFT", saoFrame, "BOTTOMRIGHT", 0, 0);
	elseif ( position == "BOTTOMLEFT" ) then
		width, height = shortSide, shortSide;
		overlay:SetPoint("TOPRIGHT", saoFrame, "BOTTOMLEFT", 0, 0);
	else
		--GMError("Unknown SpellActivationOverlay position: "..tostring(position));
		return;
	end
	
	overlay:SetSize(width * scale, height * scale);
	
	overlay.texture:SetTexture(texturePath);
	overlay.texture:SetVertexColor(r / 255, g / 255, b / 255);
	
	overlay.animOut:Stop();	--In case we're in the process of animating this out.
	PlaySoundKitID(23287);
	overlay:Show();
end 

function core:HideOverlay(index, spellID)
	for i,v in pairs(self.overlayList) do
		if (i.esaIndex == index and i.esaSpellID == spellID) then
			i.pulse:Pause();
			i.animOut:Play();
			self.overlayList[i] = false;
		end
	end
end

function core:GetUnusedOverlay()
	for i,v in pairs(self.overlayList) do
		if (not v) then
			return i;
		end
	end
	
	return self:CreateOverlay();
end

function core:CreateOverlay()
	return CreateFrame("Frame", nil, saoFrame, "SpellActivationOverlayTemplate");
end

function core:OnFadeOutFinished(anim)
	local overlay = anim:GetRegionParent();
	local overlayParent = overlay:GetParent();
	overlay.pulse:Stop();
	overlay:Hide();
end

-- End blizzard re-writes

function core:CheckTime(spellID)
	if (not self.db.profile.warn.enabled) then
		return
	end
	
	local expires = select(7, self:UnitAura(GetSpellInfo(spellID)));
	if (expires) then
		if (expires-GetTime() < self.db.profile.warn.thresh+0.1) then
			-- re-show as red
			local colors = self.db.profile.warn.color;
			self.buffStatus[spellID].redFlag = true;
			SpellActivationOverlay_ShowAllOverlays(saoFrame, spellID, dbBuff[spellID].texture, dbBuff[spellID].positions, dbBuff[spellID].scale, colors.r*255, colors.g*255, colors.b*255);
		else
			-- spell was refreshed, reschedule
			self:ScheduleTimer(function()
				self:CheckTime(spellID);
			end, expires-GetTime()-self.db.profile.warn.thresh);
		end	
	end
end

function core:UnitAura(aura)
	for i=1,80 do
		local name = UnitBuff("player", i);
		
		if (not name) then
			return nil;
		elseif (name:find(aura)) then
			return UnitBuff("player", i);
		end
	end
	
	return nil;
end

-- Blizzard function raw hook
function core:SpellActivationOverlay_ShowAllOverlays(...)
	local _, spellID = ...;
	if (spellID and self.db.profile.toggle[spellID] == false) then
		return false
	end
	
	return self.hooks["SpellActivationOverlay_ShowAllOverlays"](...);
end

function core:Debug(level, ...)
	if (type(level) ~= "number") then
		self:Debug(1, level, ...);
		return;
	end
	if self.debug >= level then 
		self:Print(...);
	end
end